ModelMap 
========

[![Build Status](https://drone.io/github.com/parnham/model_map.dart/status.png)](https://github.com/parnham/model_map.dart/latest)

An experiment in using mirrors to convert from a map to a model and vice versa.


Introduction
------------

Mirrors are a way of performing reflection in Dart. This permits an unknown class to
be traversed and information about fields, properties and methods to be extracted, which
is particularly useful if you wish to convert a serialized representation of an object 
such as JSON into an actual object instance. 

Be warned that mirrors are still being developed and so leading up to the release of 
Dart v1.0 there may still be breaking changes that would stop this library from functioning.

Anything based on mirrors can not yet be fully compiled to javascript and so ModelMap is
not recommended for browser application development at the moment.

At this stage ModelMap only worries about non-static, public fields and will ignore getters 
and setters.


It's all about the future
-------------------------

Since much of the mirror functionality is asynchronous it therefore relies on futures 
to provide results. Dart does not support blocking to await future completion and so
ModelMap cannot return results synchronously but must also return a future.

Internally there is a fair amount of recursive future handling going on since it can
only process models within models when the mirror system provides results. This caused
quite a headache when developing, with various stages of the object parsing completing
out of order, which was only solved once I had a better understanding of futures.

Fortunately the arrays of futures and sub-futures are hidden away by ModelMap behind a
single future which is quite easy to use.


Examples
--------

### Simple model

```dart
import 'package:model_map/model_map.dart';

class SimpleModel extends ModelMap
{
  String string;
  int integer;
  bool flag;
  num float;
}

main() 
{
  var map = { 'string': 'some text', 'integer': 42, 'flag': true, 'float': 1.23 };
  
  new SimpleModel().fromMap(map).then((SimpleModel model) {
    // The model is populated and ready to use at this point
  });
}
```

You can also take an existing model and convert it to a map

```dart
model.toMap().then((Map map) {
  // A map of <String, dynamic> has been created
});
```


### Using JSON

A couple of utility functions are included that simply wrap fromMap and toMap 
using the built-in parse and stringify capabilities so that you can simply call
fromJson and toJson on your model instance.


### Complex model

ModelMap can support model instances within models, but they must all extend
the ModelMap class. It also handles List and Map collections (maps are limited
to string keys only, since this is all that JSON can use). 

A limitation of Dart, which may change in the future, is that complex fields
cannot be assigned using mirrors (complex is anything that is not String, int, 
double, num or bool). To allow ModelMap to populate an entire object tree you
must assign instances to inner models, lists and maps.

```dart
import 'package:model_map/model_map.dart';

class ComplexModel extends ModelMap
{
  int integer;
  SimpleModel simple                          = new SimpleModel();
  List<SimpleModel> modelList                 = new List<SimpleModel>();
  Map<String, Date> dateMap                   = new Map<String, Date>();
  Map<String, SimpleModel> modelMap           = new Map<String, SimpleModel>();
  Map<String, List<SimpleModel>> modelListMap = new Map<String, List<SimpleModel>>();
}

main()
{
  var json = getJsonFromServer();
  
  new ComplexModel().fromJson(json).then((ComplexModel model) {
    print(model.modelMap['a key'].flag);
  });
}
```

ModelMap should cope with reasonably complex object trees, as shown in the example above, 
as long as complex fields are initialised with empty instances.


### Handling a Date

Date objects pose a problem, they are classed as complex and can therefore not be assigned
using mirrors. But they are also immutable, which means you cannot initialise a Date field
with an instance and have ModelMap modify the value.

The way to handle a Date field in ModelMap is to register a setter.

```dart

// This will not work
class DateModel extends ModelMap
{
  Date date;
}

// This will work
class DateModelWithSetter extends ModelMap
{
  Date date;


  DateModelWithSetter()
  {
    register('date', (v) => this.date = v);
  }
}
```

The register function takes the field name and a FieldSetter function which can perform the
actual assignment. In fact, any field assignment can be overridden with register thereby 
extending what ModelMap can cope with.

The setter function would normally take a Map<String, dynamic> which is a sub-section of the
complete map being parsed, but Date fields are a special case and ModelMap will actually pass
a properly initialised Date instance to the setter.

When converting to JSON or a map, ModelMap will always convert dates to an ISO 8601 string, 
however when parsing a map it will accept an ISO 8601 string or an integer representing
UTC unix time.
